import pickle
from abc import ABC
from twenty48.GameInformation import GameInformation


class Encoder(ABC):

    @staticmethod
    def encode(gameinfo: GameInformation, filename: str) -> None:
        pass


class PickleEncoder(Encoder):

    @staticmethod
    def encode(gameinfo: GameInformation, filename: str) -> None:
        with open(filename, 'ab') as file:
            pickle.dump(gameinfo, file)




from twenty48.GameInformation import GameInformation
from twenty48.Display import Display


class GameReplay:

    def __init__(self, display:Display):
        self.display = display

    def replay(self, game: GameInformation):
        turns = game.turns

        for turn in game.turns:
            self.display.display_board(turn.board.get_board_from_text_rep())
            input()



import pickle
from abc import ABC
from twenty48.GameInformation import GameInformation
from typing import List


class Decoder(ABC):
    @staticmethod
    def decode(filename: str) -> GameInformation:
        pass


class PickleDecoder(Decoder):

    @staticmethod
    def decode(filename: str) -> List[GameInformation]:
        info = []
        with open(filename, 'rb') as file:
            while True:
                try:
                    info.append(pickle.load(file))
                except EOFError:
                    break
        return info




import time
from abc import ABC, abstractmethod
import multiprocessing as mp
import numpy as np

from datetime import datetime
from typing import List
from twenty48tools.Encoder import Encoder
from twenty48.Game import Game
from twenty48.Input import Input
from twenty48.Display import NoneDisplay, ProgressDisplay


class Reporter(ABC):

    @abstractmethod
    def __init__(self, input_:Input):
        pass

    @abstractmethod
    def generate_report(self, num_games: int):
        pass


class FileReporter(Reporter):
    # Plays n games and saves gameinfo objects as a binary file for efficient storage

    def __init__(self, input_:Input, encoder: Encoder, filename: str = None):
        self.input = input_
        self.display = NoneDisplay()
        self.encoder = encoder
        if filename is None:
            self.filename = str(datetime.now()) + ".txt"
        else:
            self.filename = filename

    def generate_report(self, num_games: int = 5):
        t = time.perf_counter()
        print(f"Saving to {self.filename}")
        for i in range(num_games):
            g = Game(display=self.display, input=self.input)
            info = g.play_game()
            self.encoder.encode(gameinfo=info, filename=self.filename)
            print(f"Game number {i+1} completed", end='\r')
        t = time.perf_counter() - t
        print(f"All games completed in {t} seconds")

class LightConcurrentReporter(Reporter):
    # plays n concurrent games and reports the quartiles, mean and std.dev of mean 
    def __init__(self,  input_:Input, threads:int = 10, filename: str = None):
        self.input = input_
        self.num_threads = threads
        self.display = NoneDisplay()
        self.results = mp.Queue()
        if filename is None:
            self.filename = str(datetime.now()) + ".txt"
        else:
            self.filename = filename


    def generate_report(self, num_games: int):
        processes = [mp.Process(target=self.play_game, args=(1, self.results)) for x in range(self.num_threads)]

        for p in processes:
            p.start()

        for p in processes:
            p.join()

        results = [self.results.get() for p in processes]

        with open(self.filename, 'ab') as file:
            file.write(np.mean(results), np.std(results))
        


    def play_game(self, num_games: int, output):
        for _ in range(num_games):
            g = Game(display=self.display, input=self.input)
            output.put(g.play_game().score)

        






from twenty48.GameInformation import GameInformation
from typing import List


class Statistics:

    @staticmethod
    def avg_score(games: List[GameInformation]):
        scores = []
        for game in games:
            scores.append(game.score)
        return float(sum(scores))/float(len(scores))


